#!/usr/bin/env python

"""
Moin wiki parsing functionality.

Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.tree import Block, Region, Text
import re

# Pattern management.

def get_patterns(syntax):

    "Define patterns for the regular expressions in the 'syntax' mapping."

    patterns = {}
    for name, value in syntax.items():
        patterns[name] = re.compile(value, re.UNICODE | re.MULTILINE)
    return patterns

def combine_patterns(patterns, syntax):

    "Combine 'patterns' with those defined by the given 'syntax' mapping."

    p = {}
    p.update(patterns)
    p.update(get_patterns(syntax))
    return p

# Tokenising functions.

class TokenStream:

    "A stream of tokens taken from a string."

    def __init__(self, s, patterns):
        self.s = s
        self.patterns = patterns
        self.pos = 0
        self.match = None
        self.matching = None

    def rewind(self, length):

        "Rewind in the string by 'length'."

        self.pos -= min(length, self.pos)

    def read_until(self, pattern_names, remaining=True):

        """
        Find the first match for the given 'pattern_names'. Return the text
        preceding any match, the remaining text if no match was found, or None
        if no match was found and 'remaining' is given as a false value.
        """

        first = None
        self.matching = None

        # Find the first matching pattern.

        for pattern_name in pattern_names:
            match = self.patterns[pattern_name].search(self.s, self.pos)
            if match:
                start, end = match.span()
                if self.matching is None or start < first:
                    first = start
                    self.matching = pattern_name
                    self.match = match

        if self.matching is None:
            if remaining:
                return self.s[self.pos:]
            else:
                return None
        else:
            return self.s[self.pos:first]

    def read_match(self, group=1):

        """
        Return the matched text, updating the position in the stream. If 'group'
        is specified, the indicated group in a match will be returned.
        Typically, group 1 should contain all pertinent data, but groups defined
        within group 1 can provide sections of the data.
        """

        if self.match:
            _start, self.pos = self.match.span()
            try:
                return self.match.group(group)
            except IndexError:
                return ""
        else:
            self.pos = len(self.s)
            return None



# Utility functions.

def new_block(region):

    "Start a new block in 'region'."

    region.add(Block([]))



# Parser abstractions.

class ParserBase:

    "Common parsing methods."

    def __init__(self, formats=None):

        """
        Initialise the parser with any given 'formats' mapping from region type
        names to parser objects.
        """

        self.formats = formats

    def get_items(self, s):

        "Return a sequence of token items for 's'."

        raise NotImplementedError

    def parse(self, s):

        """
        Parse page text 's'. Pages consist of regions delimited by markers.
        """

        return self.parse_region(self.get_items(s))

    def parse_region(self, items, level=0, indent=0):

        """
        Parse the data provided by 'items' to populate a region with the given
        'level' at the given 'indent'.
        """

        region = Region([], level, indent)

        # Parse section headers, then parse according to region type.

        self.parse_region_header(items, region)
        self.parse_region_type(items, region)

        return region

    def parse_region_type(self, items, region):

        """
        Given data provided by 'items', use configured parsers to parse the
        'region' based on its type.
        """

        # Find an appropriate parser given the type.

        if self.formats.has_key(region.type):
            self.formats[region.type].parse_region_content(items, region)

        # Otherwise, treat the section as opaque.

        else:
            self.parse_region_opaque(items, region)

    def parse_region_header(self, items, region):

        """
        Parse the region header from the 'items', setting it for the given 'region'.
        """

        if items.read_until(["header"], False) == "": # None means no header
            region.type = items.read_match()

    def parse_region_opaque(self, items, region):

        "Parse the data provided by 'items' to populate an opaque 'region'."

        region.transparent = False
        self.parse_region_details(items, region, ["regionend"])

    def parse_region_content(self, items, region):

        "Parse the data provided by 'items' to populate the given 'region'."

        pass

    # Parsing utilities.

    def parse_region_details(self, items, region, pattern_names):

        "Parse 'items' within 'region' searching using 'pattern_names'."

        try:
            while True:

                # Obtain text before any marker or the end of the input.

                preceding = items.read_until(pattern_names)
                if preceding:
                    region.append_inline(Text(preceding))

                # End of input.

                if not items.matching:
                    break

                # Obtain any feature.

                feature = items.read_match()
                handler = self.handlers.get(items.matching)

                # Handle each feature or add text to the region.

                if handler:
                    handler(self, items, region)
                else:
                    region.append_inline(Text(feature))

        except StopIteration:
            pass

        region.normalise()

    def end_region(self, items, region):

        "End the parsing of 'region', breaking out of the parsing loop."

        raise StopIteration

# vim: tabstop=4 expandtab shiftwidth=4
